"""This module contains Textual widgets for HumBLE Explorer's user interface."""
from __future__ import annotations

from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.widgets import Checkbox, Input, Static

__author__ = "Koen Vervloesem"
__copyright__ = "Koen Vervloesem"
__license__ = "MIT"


class FilterWidget(Input):
    """A Textual widget to filter Bluetooth Low Energy advertisements."""

    def __init__(self, placeholder: str = "") -> None:
        """Create new FilterWidget.

        Args:
            placeholder (str): Placeholder to show in the filter widget.
        """
        super().__init__(placeholder=placeholder)
        self.display = False

    def on_blur(self) -> None:
        """Automatically hide widget on losing focus."""
        self.display = False


class SettingsWidget(Static):
    """A Textual widget to let the user choose settings."""

    def __init__(self, id: str | None) -> None:
        """Create new SettingsWidget.

        Args:
            id (str): Id of the settings widget.
        """
        super().__init__(id=id)
        self.display = False

    def compose(self) -> ComposeResult:
        """Show checkboxes."""
        yield Static("[b]Show data types[/b]\n")
        yield Horizontal(
            Static("Local name       ", classes="label"),
            Checkbox(value=True, id="local_name", classes="view"),
            classes="container",
        )
        yield Horizontal(
            Static("RSSI             ", classes="label"),
            Checkbox(value=True, id="rssi", classes="view"),
            classes="container",
        )
        yield Horizontal(
            Static("TX power         ", classes="label"),
            Checkbox(value=True, id="tx_power", classes="view"),
            classes="container",
        )
        yield Horizontal(
            Static("Manufacturer data", classes="label"),
            Checkbox(value=True, id="manufacturer_data", classes="view"),
            classes="container",
        )
        yield Horizontal(
            Static("Service data     ", classes="label"),
            Checkbox(value=True, id="service_data", classes="view"),
            classes="container",
        )
        yield Horizontal(
            Static("Service UUIDs    ", classes="label"),
            Checkbox(value=True, id="service_uuids", classes="view"),
            classes="container",
        )
        yield Static("\n[b]Other settings[/b]\n")
        yield Horizontal(
            Static("Auto-scroll      ", classes="label"),
            Checkbox(value=True, id="autoscroll", classes="view"),
            classes="container",
        )

    def on_blur(self) -> None:
        """Automatically hide widget on losing focus."""
        self.display = False



"""Main entry point for HumBLE Explorer."""
from __future__ import annotations

import sys
from argparse import ArgumentParser, Namespace

from humble_explorer import __version__
from humble_explorer.app import BLEScannerApp

__author__ = "Koen Vervloesem"
__copyright__ = "Koen Vervloesem"
__license__ = "MIT"


# ---- CLI ----
# The functions defined in this section are wrappers around the main Python
# API allowing them to be called directly from the terminal as a CLI
# executable/script.


def run() -> None:
    """Calls :func:`main` passing the CLI arguments extracted from `sys.argv`.

    This function can be used as entry point to create console scripts with setuptools.
    """
    main(sys.argv[1:])


def main(args: list[str]) -> None:
    """Wrapper allowing the BLE scanner app to be called on the command line.

    This wrapper accepts string arguments.

    Args:
      args (list[str]): command line parameters as list of strings
          (for example  ``["--scanning-mode", "passive"]``).
    """
    cli_args = parse_args(args)
    app = BLEScannerApp(cli_args=cli_args)
    app.run()


def parse_args(args: list[str]) -> Namespace:
    """Parse command line parameters.

    Args:
      args (list[str]): command line parameters as list of strings
          (for example  ``["--help"]``).

    Returns:
      `argparse.Namespace`: command line parameters namespace
    """
    parser = ArgumentParser(description="Human-friendly Bluetooth Low Energy Explorer")
    parser.add_argument(
        "--version",
        action="version",
        version=f"humble-explorer {__version__}",
    )
    parser.add_argument(
        "-a", "--adapter", help="Bluetooth adapter (e.g. hci1 on Linux)", type=str
    )
    parser.add_argument(
        "-s",
        "--scanning-mode",
        dest="scanning_mode",
        help="Scanning mode (default: active)",
        type=str,
        default="active",
        choices=("active", "passive"),
    )
    return parser.parse_args(args)


if __name__ == "__main__":
    # ^  This is a guard statement that will prevent the following code from
    #    being executed in the case someone imports this file instead of
    #    executing it as a script.
    #    https://docs.python.org/3/library/__main__.html
    run()



"""Module with Rich renderables for HumBLE Explorer's user interface."""
from __future__ import annotations

from datetime import datetime
from string import printable, whitespace
from uuid import UUID

from bleak.backends.scanner import AdvertisementData
from bluetooth_numbers import company, service
from bluetooth_numbers.exceptions import UnknownCICError, UnknownUUIDError
from rich._palettes import EIGHT_BIT_PALETTE
from rich.style import Style
from rich.table import Table
from rich.text import Text
from rich.tree import Tree

from humble_explorer.utils import hash8

__author__ = "Koen Vervloesem"
__copyright__ = "Koen Vervloesem"
__license__ = "MIT"

PRINTABLE_CHARS = printable.replace(whitespace, " ")


class RichTime:
    """Rich renderable that shows a time.

    All times within the same second are rendered in the same color.
    """

    def __init__(self, time: datetime, style: Style | None = None) -> None:
        """Create a RichTime object.

        Args:
            time (datetime): The time to show.
            style (Style): The Rich style to use.
        """
        self.full_time = time.strftime("%H:%M:%S.%f")
        if style:
            self.style = style
        else:
            self.style = Style(
                color=EIGHT_BIT_PALETTE[hash8(time.strftime("%H:%M:%S"))].hex
            )

    def __rich__(self) -> Text:
        """Render the RichTime object.

        Returns:
            Text: The rendering of the RichTime object.
        """
        return Text(self.full_time, style=self.style)


class RichDeviceAddress:
    """Rich renderable that shows a Bluetooth device address.

    Every address is rendered in its own color.
    """

    def __init__(self, address: str) -> None:
        """Create a RichDeviceAddress object.

        Args:
            address (str): The address to show.
        """
        self.address = address
        self.style = Style(color=EIGHT_BIT_PALETTE[hash8(self.address)].hex)

    def __rich__(self) -> Text:
        """Render the RichDeviceAddress object.

        Returns:
            Text: The rendering of the RichDeviceAddress object.
        """
        return Text(self.address, style=self.style)


class RichRSSI:
    """Rich renderable that shows RSSI of a device."""

    def __init__(self, rssi: int) -> None:
        """Create a RichRSSI object.

        Args:
            rssi (int): The RSSI to show.
        """
        self.rssi = rssi

    def __rich__(self) -> Text:
        """Render the RichRSSI object.

        Returns:
            Text: The rendering of the RichRSSI object.
        """
        return Text.assemble((str(self.rssi), "green bold"), " dBm")


class RichUUID:
    """Rich renderable that shows a UUID with description and colors."""

    def __init__(self, uuid128: str) -> None:
        """Create a RichUUID object.

        Args:
            uuid128 (str): The UUID to show.
        """
        self.uuid128 = uuid128

    def __rich__(self) -> Text:
        """Render the RichUUID object.

        Returns:
            Text: The rendering of the RichUUID object.
        """
        # Colorize the 16-bit UUID part in a standardized 128-bit UUID.
        if self.uuid128.startswith("0000") and self.uuid128.endswith(
            "-0000-1000-8000-00805f9b34fb"
        ):
            colored_uuid = Text.assemble(
                "0000",
                (self.uuid128[4:8], "green bold"),
                "-0000-1000-8000-00805f9b34fb",
            )
        else:
            colored_uuid = Text(self.uuid128)

        try:
            return Text.assemble(colored_uuid, f" ({service[UUID(self.uuid128)]})")
        except UnknownUUIDError:
            return Text.assemble(colored_uuid, " (Unknown)")


class RichCompanyID:
    """Rich renderable that shows company ID and name."""

    def __init__(self, cic: int) -> None:
        """Create a RichCompanyID object.

        Args:
            cic (int): The company ID to show.
        """
        self.cic = cic

    def __rich__(self) -> Text:
        """Render the RichCompanyID object.

        Returns:
            Text: The rendering of the RichCompanyID object.
        """
        try:
            manufacturer_name = company[self.cic]
        except UnknownCICError:
            manufacturer_name = "Unknown"

        return Text.assemble(
            f"0x{self.cic:04x} (", (manufacturer_name, "green bold"), ")"
        )


class RichHexData:
    """Rich renderable that shows hex data."""

    def __init__(self, data: bytes) -> None:
        """Create a RichHexData object.

        Args:
            data (bytes): The hex data to show.
        """
        self.data = data

    def __rich__(self) -> Text:
        """Render the RichHexData object.

        Returns:
            Text: The rendering of the RichHexData object.
        """
        # Python 3.7 doesn't have sep parameter for bytes.hex()
        hex_data_str = self.data.hex()
        hex_data = " ".join(
            a + b for a, b in zip(hex_data_str[::2], hex_data_str[1::2])
        )
        return Text(hex_data, style="cyan bold")


class RichHexString:
    """Rich renderable that shows hex data as a string.

    Non-printable characters are replaced by a dot.
    """

    def __init__(self, data: bytes) -> None:
        """Create a RichHexString object.

        Args:
            data (bytes): The hex data to show.
        """
        self.data = data

    def __rich__(self) -> str:
        """Render the RichHexString object.

        Returns:
            Text: The rendering of the RichHexString object.
        """
        result = []
        for byte in self.data:
            char = chr(byte)
            if char in PRINTABLE_CHARS:
                result.append(f" {char}")
            else:
                result.append(" .")

        return " ".join(result)


class RichAdvertisement:
    """Rich renderable that shows advertisement data."""

    def __init__(self, data: AdvertisementData, show_data: dict[str, bool]) -> None:
        """Create a RichAdvertisement object.

        Args:
            data (AdvertisementData): The advertisement data to show.
            show_data (dict[str, bool]): Which data to show.
        """
        self.data = data
        self.show_data = show_data

    def height(self) -> int:
        """Return the number of lines this Rich renderable uses."""
        height = 0
        if self.data.local_name and self.show_data["local_name"]:
            height += 1
        if self.data.rssi and self.show_data["rssi"]:
            height += 1
        if self.data.tx_power and self.show_data["tx_power"]:
            height += 1
        if self.data.manufacturer_data and self.show_data["manufacturer_data"]:
            height = height + 1 + 3 * len(self.data.manufacturer_data)
        if self.data.service_data and self.show_data["service_data"]:
            height = height + 1 + 3 * len(self.data.service_data)
        if self.data.service_uuids and self.show_data["service_uuids"]:
            height = height + 1 + len(self.data.service_uuids)
        return height  # noqa: R504

    def __rich__(self) -> Table:
        """Render the RichAdvertisement object.

        Returns:
            Table: The rendering of the RichAdvertisement object.
        """
        table = Table(show_header=False, show_edge=False, padding=0)

        # Show local name
        if self.data.local_name and self.show_data["local_name"]:
            table.add_row(
                Text.assemble("local name: ", (self.data.local_name, "green bold"))
            )

        # Show RSSI
        if self.data.rssi and self.show_data["rssi"]:
            table.add_row(Text.assemble("RSSI: ", RichRSSI(self.data.rssi).__rich__()))

        # Show TX Power
        if self.data.tx_power and self.show_data["tx_power"]:
            table.add_row(
                Text.assemble("TX power: ", RichRSSI(self.data.tx_power).__rich__())
            )

        # Show manufacturer data
        if self.data.manufacturer_data and self.show_data["manufacturer_data"]:
            tree = Tree("manufacturer data:")
            for cic, value in self.data.manufacturer_data.items():
                company_structure = Tree(
                    Text.assemble(
                        RichCompanyID(cic).__rich__(), f" → {len(value)} bytes"
                    )
                )
                company_structure.add(
                    Text.assemble("hex  → ", RichHexData(value).__rich__())
                )
                company_structure.add(
                    Text.assemble("text → ", RichHexString(value).__rich__())
                )
                tree.add(company_structure)
            table.add_row(tree)

        # Show service data
        if self.data.service_data and self.show_data["service_data"]:
            tree = Tree("service data:")
            for uuid, value in self.data.service_data.items():
                svc_uuid = Tree(
                    Text.assemble(RichUUID(uuid).__rich__(), f" → {len(value)} bytes")
                )
                svc_uuid.add(Text.assemble("hex  → ", RichHexData(value).__rich__()))
                svc_uuid.add(Text.assemble("text → ", RichHexString(value).__rich__()))
                tree.add(svc_uuid)
            table.add_row(tree)

        # Show service UUIDs with their description
        if self.data.service_uuids and self.show_data["service_uuids"]:
            tree = Tree("service UUIDs:")
            for uuid in sorted(self.data.service_uuids):
                tree.add(RichUUID(uuid))
            table.add_row(tree)

        return table



"""Tests for renderables module."""
from datetime import datetime

from rich.text import Span, Text

from humble_explorer.renderables import (
    RichCompanyID,
    RichDeviceAddress,
    RichHexData,
    RichHexString,
    RichRSSI,
    RichTime,
    RichUUID,
)

__author__ = "Koen Vervloesem"
__copyright__ = "Koen Vervloesem"
__license__ = "MIT"


def test_time() -> None:
    """Test RichTime class."""
    time1 = RichTime(datetime.now())
    time2 = RichTime(datetime.now())
    # Make time1 and time2 times within the same second
    while time1.full_time.split(".")[0] != time2.full_time.split(".")[0]:
        time1 = RichTime(datetime.now())

    # Times within the same second should have the same color
    assert time1.style == time2.style


def test_device_address() -> None:
    """Test RichDeviceAddress class."""
    address_string = "D5:FE:15:49:AC:7D"
    device_address = RichDeviceAddress(address_string)
    device_address2 = RichDeviceAddress(address_string)
    assert str(device_address.__rich__()) == address_string

    # Devices with the same address should have the same color
    assert device_address.style == device_address2.style


def test_rssi() -> None:
    """Test RichRSSI class."""
    rssi = RichRSSI(-70)

    # The number should be colored
    assert rssi.__rich__() == Text.assemble(("-70", "green bold"), " dBm")


def test_uuid() -> None:
    """Test RichUUID class."""
    # Unknown UUID should be rendered without color
    unknown_uuid = "22110000-554a-4546-5542-46534450464d"
    assert str(RichUUID(unknown_uuid).__rich__()) == f"{unknown_uuid} (Unknown)"

    # 16-bit UUID part in a standardized 128-bit UUID should be colored
    environmental_sensing_uuid = "0000181a-0000-1000-8000-00805f9b34fb"
    assert (
        str(RichUUID(environmental_sensing_uuid).__rich__())
        == "0000181a-0000-1000-8000-00805f9b34fb (Environmental Sensing)"
    )
    assert (
        Span(4, 8, "green bold")
        in RichUUID(environmental_sensing_uuid).__rich__().spans
    )


def test_company_id() -> None:
    """Test RichCompanyID class."""
    # Company name should be colored
    ruuvi_cid = 0x0499
    assert str(RichCompanyID(ruuvi_cid).__rich__()) == "0x0499 (Ruuvi Innovations Ltd.)"
    assert (
        Span(8, len(str(RichCompanyID(ruuvi_cid).__rich__())) - 1, "green bold")
        in RichCompanyID(ruuvi_cid).__rich__().spans
    )


def test_hex_data() -> None:
    """Test RichHexData class."""
    # Hex data should be colored and with space separator between bytes
    data = b"\x03\x4e\x11\x23\xca\x2e\xff\xaa\xff\xd1\x03\xe7\x0b\xb9"
    assert (
        str(RichHexData(data).__rich__()) == "03 4e 11 23 ca 2e ff aa ff d1 03 e7 0b b9"
    )
    assert RichHexData(data).__rich__().style == "cyan bold"


def test_hex_string() -> None:
    """Test RichHexString class."""
    # Hex string should show the correct characters with space separator between bytes
    data = b"\x03\x4e\x11\x23\xca\x2e\xff\xaa\xff\xd1\x03\xe7\x0b\xb9"
    assert (
        str(RichHexString(data).__rich__())
        == " .  N  .  #  .  .  .  .  .  .  .  .  .  ."
    )



"""Module with the Textual app that scans for Bluetooth Low Energy advertisements."""
from __future__ import annotations

from argparse import Namespace
from datetime import datetime
from platform import system

from bleak import BleakScanner
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData
from rich.style import Style
from textual.app import App, ComposeResult
from textual.reactive import reactive
from textual.widgets import Checkbox, DataTable, Footer, Header, Input

if system() == "Linux":
    from bleak.assigned_numbers import AdvertisementDataType
    from bleak.backends.bluezdbus.advertisement_monitor import OrPattern
    from bleak.backends.bluezdbus.scanner import BlueZScannerArgs

from humble_explorer.renderables import RichAdvertisement, RichDeviceAddress, RichTime
from humble_explorer.widgets import FilterWidget, SettingsWidget

from . import __version__

__author__ = "Koen Vervloesem"
__copyright__ = "Koen Vervloesem"
__license__ = "MIT"

_PAUSE_STYLE = Style(color="red", bgcolor="grey50")


class BLEScannerApp(App[None]):
    """A Textual app to scan for Bluetooth Low Energy advertisements."""

    CSS_PATH = "app.css"

    BINDINGS = [
        ("q", "quit", "Quit"),
        ("f", "toggle_filter", "Filter"),
        ("s", "toggle_settings", "Settings"),
        ("t", "toggle_scan", "Toggle scan"),
        ("c", "clear_advertisements", "Clear"),
    ]

    address_filter = reactive("")  #: :meta private:

    def __init__(self, cli_args: Namespace) -> None:
        """Initialize BLE scanner.

        Args:
            cli_args (argparse.Namespace): Command-line arguments.
        """
        # Configure scanning mode
        self.scanner_kwargs = {"scanning_mode": cli_args.scanning_mode}

        if system() == "Linux":
            if cli_args.scanning_mode == "passive":
                # Passive scanning with BlueZ needs at least one or_pattern.
                # The following matches all devices.
                self.scanner_kwargs["bluez"] = BlueZScannerArgs(
                    or_patterns=[
                        OrPattern(0, AdvertisementDataType.FLAGS, b"\x06"),
                        OrPattern(0, AdvertisementDataType.FLAGS, b"\x1a"),
                    ]
                )
            elif cli_args.scanning_mode == "active":
                # Disable duplicate detection of advertisement data
                # for a more low-level view of what packets are really sent.
                self.scanner_kwargs["bluez"] = BlueZScannerArgs(
                    filters={"DuplicateData": True}
                )

        # Configure Bluetooth adapter
        self.scanner_kwargs["adapter"] = cli_args.adapter

        # Configure scanner
        self.scanner_kwargs["detection_callback"] = self.on_advertisement
        self.scanning = False

        # Initialize empty list of advertisements
        self.advertisements: list[tuple[datetime, str, AdvertisementData]] = []

        super().__init__()

    def set_title(self, scanning_description: str) -> None:
        """Set the title of the app with a description of the scanning status.

        Args:
            scanning_description (str): Description of the scanning status.
        """
        self.title = f"HumBLE Explorer {__version__} ({scanning_description})"

    def action_toggle_settings(self) -> None:
        """Enable or disable settings widget."""
        settings_widget = self.query_one(SettingsWidget)
        settings_widget.display = not settings_widget.display

    def action_toggle_filter(self) -> None:
        """Enable or disable filter input widget."""
        filter_widget = self.query_one(FilterWidget)
        filter_widget.display = not filter_widget.display
        if filter_widget.display:
            self.set_focus(filter_widget)

    async def action_toggle_scan(self) -> None:
        """Start or stop BLE scanning."""
        if self.scanning:
            await self.stop_scan()
        else:
            await self.start_scan()

    def action_clear_advertisements(self) -> None:
        """Clear the list of received advertisements."""
        self.advertisements = []
        self.query_one(DataTable).clear()

    def compose(self) -> ComposeResult:
        """Create child widgets for the app.

        Returns:
            textual.app.ComposeResult: The child widgets for the app.
        """
        yield Header()
        yield Footer()
        yield SettingsWidget(id="sidebar")
        yield FilterWidget(placeholder="address=")
        yield DataTable(zebra_stripes=True)

    def show_data_config(self) -> dict[str, bool]:
        """Return dictionary with which advertisement data to show.

        Returns:
            dict[str, bool]: Each key has the value ``True`` if this advertisement
                type should be shown and ``False`` if not.
        """
        return {
            "local_name": self.query_one("#local_name", Checkbox).value,
            "rssi": self.query_one("#rssi", Checkbox).value,
            "tx_power": self.query_one("#tx_power", Checkbox).value,
            "manufacturer_data": self.query_one("#manufacturer_data", Checkbox).value,
            "service_data": self.query_one("#service_data", Checkbox).value,
            "service_uuids": self.query_one("#service_uuids", Checkbox).value,
        }

    async def on_advertisement(
        self, device: BLEDevice, advertisement_data: AdvertisementData
    ) -> None:
        """Show advertisement data on detection of a BLE advertisement.

        Args:
            device (~bleak.backends.device.BLEDevice): The device advertising the data.
            advertisement_data (~bleak.backends.scanner.AdvertisementData): The
                advertised data.
        """
        self.log(advertisement_data.local_name, device.address, advertisement_data)

        # Append advertisement to list of all advertisements
        now = datetime.now()
        self.advertisements.append((now, device.address, advertisement_data))

        # Create renderables for advertisement and add them to table
        table = self.query_one(DataTable)
        self.add_advertisement_to_table(
            table,
            RichTime(now),
            RichDeviceAddress(device.address),
            RichAdvertisement(advertisement_data, self.show_data_config()),
        )

    async def on_mount(self) -> None:
        """Initialize interface and start BLE scan."""
        table = self.query_one(DataTable)
        table.add_columns("Time", "Address", "Advertisement")
        # Set focus to table for immediate keyboard navigation
        table.focus()

        # Set up Bleak scanner and start BLE scan
        self.scanner = BleakScanner(**self.scanner_kwargs)
        await self.start_scan()

    def on_checkbox_changed(self, message: Checkbox.Changed) -> None:
        """React when the checkbox is ticked or unticked.

        Show or hide advertisement data depending on the state of
        the checkboxes.

        Args:
            message (textual.widgets.Checkbox.Changed): The message with the changed
                checkbox.
        """
        if "view" in message.input.classes:
            self.recreate_table()

    def on_input_changed(self, message: Input.Changed) -> None:
        """Filter advertisements with user-supplied filter.

        Args:
            message (textual.widgets.Input.Changed): The message with the user's
                changed input.
        """
        if message.value.startswith("address="):
            self.address_filter = message.value[8:].upper()
        else:
            self.address_filter = ""

    def watch_address_filter(self, old_filter: str, new_filter: str) -> None:
        """React when the reactive attribute address_filter changes.

        This recreates the table.

        Args:
            old_filter (str): The old value of the filter.
            new_filter (str): The new value of the filter.
        """
        self.recreate_table()

    def recreate_table(self) -> None:
        """Recreate table with advertisements."""
        table = self.query_one(DataTable)
        table.clear()
        for advertisement in self.advertisements:
            self.add_advertisement_to_table(
                table,
                RichTime(advertisement[0]),
                RichDeviceAddress(advertisement[1]),
                RichAdvertisement(advertisement[2], self.show_data_config()),
            )

    def scroll_if_autoscroll(self) -> None:
        """Scroll to the end if autoscroll is enabled."""
        if self.query_one("#autoscroll", Checkbox).value:
            self.query_one(DataTable).scroll_end(animate=False)

    def add_advertisement_to_table(
        self,
        table: DataTable,
        now: RichTime,
        device_address: RichDeviceAddress,
        rich_advertisement: RichAdvertisement,
    ) -> None:
        """Add new row to table with time, address and advertisement.

        Args:
            table (textual.widgets.DataTable): The table to add an advertisement to.
            now (RichTime): The time.
            device_address (RichDeviceAddress): The device address.
            rich_advertisement (RichAdvertisement): The advertisement.
        """
        if device_address.address.startswith(self.address_filter):
            table.add_row(
                now,
                device_address,
                rich_advertisement,
                height=rich_advertisement.height(),
            )
            self.scroll_if_autoscroll()

    async def start_scan(self) -> None:
        """Start BLE scan."""
        self.scanning = True
        self.set_title("Scanning")
        await self.scanner.start()

    async def stop_scan(self) -> None:
        """Stop BLE scan."""
        self.scanning = False
        self.set_title("Stopped")
        await self.scanner.stop()
        table = self.query_one(DataTable)
        table.add_row(RichTime(datetime.now(), style=_PAUSE_STYLE))
        self.scroll_if_autoscroll()



"""This module contains utility functions for HumBLE Explorer."""
from random import shuffle

__author__ = "Koen Vervloesem"
__copyright__ = "Koen Vervloesem"
__license__ = "MIT"

permutation_table = list(range(256))
shuffle(permutation_table)


def hash8(message: str) -> int:
    """Compute an 8-bit hash from the message with Pearson hashing.

    Args:
        message (str): The message to hash.

    Returns:
        int: An 8-bit hash value for `message`.
    """
    # Source: https://en.wikipedia.org/wiki/Pearson_hashing
    hash_value = len(message) % 256
    for i in message:
        hash_value = permutation_table[hash_value ^ ord(i)]
    return hash_value
